/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.TYPE		Module
.IDENTIFICATION	buffer.c
.AUTHOR    	Francois Ochsenbein [ESO-IPG]
.LANGUAGE  	C
.KEYWORDS	Buffer / Stack Management
.ENVIRONMENT 	
.COMMENTS
	These routines allow the management of buffers defined
	as the following BUFFER structure:
\begin{TeX}
\begin{itemize}
\item	{\tt char *buf} is the address of the buffer in memory
\item	{\tt int allocated}	is the number of bytes allocated to buf
\item	{\tt int used}	is the present number of bytes used in this
		buffer
\item	{\tt int increment}	is the number of bytes for automatic expansions
\item	{\tt offset}	is an integer, \eg
	to perform some searches. It is used for stacking and unstacking.
\end{itemize}
Note that {\em osm} routines (memory allocation) may be used (see buffer.h)
\end{TeX}
.VERSION 1.0	03-Feb-1987: Creation
.VERSION 1.1	20-May-1987: Removed bug in mm_set
.VERSION 2.0	20-Apr-1988: Changed function names, added stacking facilities.
.VERSION 2.1	01-Jun-1988: Added mm_zloc (retrieve a zero item),
			mm_zindex (retrieve non-zero item)
			mm_zfree (free the index)
.VERSION 2.2	07-Nov-1988: Removed bug in mm_bunst
.VERSION 2.3	07-Jun-1991: Be sure an EOS is appended in mm_bapp
.VERSION 2.4	19-Oct-1994: Cosmetics
.VERSION 2.5	06-Apr-2005: ANSI
.VERSION 2.6	02-Jun-2007: Cosmetics (-Wall option)
.VERSION 2.7	16-Sep-2010: mm_balloc accepts negative length ==> reserve only
.VERSION 2.8	11-Oct-2010: mm_printf (required C99)
.VERSION 2.9	03-Dec-2010: mm_bedi = BUF_AppendInt
.VERSION 3.0	30-May-2013: Problem when addition of a string within buffer...
----------------------------------------------------------------------------*/

#define	DEBUG		0	/* For debugging purposes	*/

#include <buffer.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define bzero(a,n)		memset(a, 0, n)
	/* Macros for memory management */
#define MEM_GET(type, n)	(type *)malloc(sizeof(type)*(n))
#define MEM_EXP(type,ptr,n)	(type *)realloc(ptr,sizeof(type)*(n))
#define MEM_FREE(ptr)		free((char *)ptr)


/*==========================================================================*/
#if DEBUG
static void mm_blist(b, text)
/*+++
.PURPOSE List (on stderr) for debugging purposes
.RETURNS Address of allocated buffer, or NULL address if failed,
.REMARKS If incr is zero, the buffer cannot be automatically expanded.
-------*/
  BUFFER *b;	/* IN: The buffer */
  char *text;	/* IN: Text	  */
{
    fprintf(stderr, "....%s: BUFFER a=%d, i=%d, u=%d, o=%d =>%p\n", 
	text, b->allocated, b->increment, b->used, b->offset, b->buf) ;
}
#endif

/*==========================================================================*/
BUFFER *mm_bnew(int size, int incr)    
/*+++
.PURPOSE Allocation of a buffer of a given size
.RETURNS Address of allocated buffer, or NULL address if failed,
.REMARKS If incr is zero, the buffer cannot be automatically expanded.
  [IN] size = The required size (bytes)
  [IN] incr = The increment size for expansion (bytes)
-------*/
{
  register BUFFER *b;

    if (size < 0) size = 0;
    b = MEM_GET(BUFFER,1);
    b->buf 	= (char *)0;
    b->allocated= size;
    b->used	= 0;
    b->increment= (incr > 0 ? incr : 0);
    b->offset	= 0;

    if (b->allocated)
  	b->buf = MEM_GET(char, b->allocated);

    return(b);
}  

/*==========================================================================*/
int mm_bfree(BUFFER *b)    
/*+++
.PURPOSE Free a BUFFER structure --- but ONLY the buffer!
.RETURNS 0
.REMARKS 
---*/
{
    if (b) {
  	MEM_FREE(b->buf);
	b->buf 		= (char *)0;
	b->allocated 	= 0;
	b->used		= 0;
	b->offset	= 0;
    }
    return(0);
}  

int mm_bkill(BUFFER *b)    
/*+++
.PURPOSE Remove completely the BUFFER structure
.RETURNS 0
.REMARKS 
---*/
{
    if (b) {
  	mm_bfree(b);
	MEM_FREE(b);
    }
    return(0);
}  

/*==========================================================================*/
int mm_bexp(BUFFER *b, int size)    
/*+++
.PURPOSE Expansion of a buffer
.RETURNS 0 (OK) / -1 (error)
.REMARKS Size of zero implies "use increment"
  [IN] b    = The Buffer to Expand
  [IN] size = The required new size (bytes)
---*/
{
  register char *p;
  int new_size, status;

    new_size = (size > 0 ? size : b->allocated + b->increment);

#if DEBUG
    fprintf(stderr, "....mm_bexp: ask for %d bytes\n", size);
    mm_blist(b, "mm_bexp: initial");
#endif

    p = MEM_EXP(char, b->buf, new_size);
    if (p)  	b->allocated = new_size, b->buf = p, status = 0;
    else	status = -1;

#if DEBUG
    fprintf(stderr, "....mm_bexp: status=%d\n", status);
    mm_blist(b, "mm_bexp: final");
#endif

    return(status);
}

/*==========================================================================*/
char *mm_bstack(BUFFER *b, char *record, int len)    
/*+++
.PURPOSE Stack a record into buffer --- The buffer is
	expanded if necessary
.RETURNS Address of stacked record / NULL if error
.REMARKS Save offset of previous stacked values.
  [IN] b       = The buffer to free
  [IN] *record = record to append	
  [IN] len     = Length of record	
---*/
{
  char	*a;
    if ((a = mm_balloc(b, len + sizeof(int)))) {
    	memcpy(a, (char *)(&(b->offset)), sizeof(int)), a += sizeof(int);
	b->offset = a - b->buf;
  	memcpy(a, record, len);
    }
    return(a);
}

/*==========================================================================*/
char *mm_bunstack(BUFFER *b)
/*+++
.PURPOSE Unstack a record from buffer.
.RETURNS Address of record / NULL
.REMARKS 
---*/
{
  char	*p;
    if (b->used <= 0)	p = (char *)0;
    else {
    	p = b->buf + b->offset;
  	b->used = b->offset - sizeof(int);
	if (b->used == 0) b->offset = 0;
	else memcpy((char *)(&(b->offset)), b->buf + b->used, sizeof(int));
    }
    return(p);
}

/*==========================================================================*/
char *mm_bapp(b, record, len)    
/*+++
.PURPOSE Append a record to a buffer --- The buffer is
	expanded if necessary
.RETURNS Address of stored record / NULL if error
.REMARKS 
---*/
  BUFFER *b;		/* IN: The buffer to free	*/
  char	*record;	/* IN: record to append		*/
  int	len;		/* IN: Length of record		*/
{
  char	*p, *a=0;
	
  /* V3.0: Problem when record included in buffer ... */
    if (b->buf && (record>=b->buf) && ((record-b->buf)<b->allocated)) {
	if ((a = MEM_GET(char, len)))
	    memcpy(a, record, len); 
	else return(a);		/* FAILED (not enough memory) */
	record = a;
    }

    /* Get a pointer in buffer, and then copy string
     * to this buffer	*/

    if ((p = mm_balloc(b, len+1))) {
  	if (len>0) memcpy(p, record, len);
	p[len] = 0;
  	b->used -= 1;	
    }
    if (a) MEM_FREE(a);

    return(p);
}  

char *mm_bedi(BUFFER *b, int n)
/*++++++++++++++++
.PURPOSE  Append a number (edited) to a buffer
.RETURNS  Base 10 edition
------------------*/
{
  char *p;
    p = mm_balloc(b, -24);  /* Reserve at least to 24 bytes; p=b->used */
    sprintf(p, "%d", n);
    b->used += strlen(p);
    return(p);
}

/*==========================================================================*/
char *mm_balloc( BUFFER *b, int len)    
/*+++
.PURPOSE Set a pointer to next available position in buffer.
	Buffer is expanded if necessary
.RETURNS Address of available record / NULL if error
.REMARKS len<0 ==> Reserve
  [IN] b   = The buffer to free
  [IN] len = Length of record
---*/
{
  char	*p;
  int	ex, neg;
	
    if (!b)	return((char *)0);
    p = (b->buf) + b->used;
    if (len==0)	return(p); 	/* V2.7: accept len<=0 */
    if (len<0) { neg = 1; len = -len; }
    else         neg = 0;

#if DEBUG
    fprintf(stderr, "....mm_balloc: ask for %d bytes\n", len);
    mm_blist(b, "mm_balloc: initial");
#endif

    ex = b->used + len - b->allocated;
    if ( ex > 0) {		/* Doesn't fit. Expand is required */
  	if(b->increment) {	/* Increment is possible */
  	    ex = (ex+b->increment-1)/b->increment;
  	    ex = ex*b->increment + b->allocated;
  	    p = (mm_bexp(b, ex) ? (char *)0 /* FAIL */ : b->buf + b->used);
	}
	else p = (char* )0;
    }
    else	;

    if(p &&(neg==0)) b->used += len;	/* Space is actually used */
    return(p);
}

/*==========================================================================*/
char *mm_zfree(BUFFER *b, int index, int item_len)    
/*+++
.PURPOSE Free the indexed item.
.RETURNS Address of freed item / NULL if bad item
.REMARKS No tracing.
  [IN] b        = The buffer where to free
  [IN] index    = Index of item to free
  [IN] item_len = Size of 1 item
---*/
{
  char	*p;
  int	o;

	/* Find first the position */
    o = index * item_len;
    if ( (o < 0) || (o >= b->used) )	return((char *)0);
    p = b->buf + o;
    bzero(p, item_len);

    return(p);
}

/*==========================================================================*/
char *mm_zindex(BUFFER *b, int index, int item_len)    
/*+++
.PURPOSE Finds the position of indexed item. Null item is checked.
.RETURNS Address of found item / NULL if bad item or is filled
	with zeroes.
.REMARKS No tracing.
  [IN] b        = The buffer to use
  [IN] index    = Index of item to locate
  [IN] item_len = Size of 1 item
---*/
{
  char	*p;
  int	o;

	/* Find first the position */
    o = index * item_len;
    if ( (o < 0) || (o >= b->used) )	return((char *)0);
    p = b->buf + o;

	/* Check the item is zero-filled */
    for (o = item_len; --o >= 0; )
	if (! p[o]) return((char *)0);
  
    return(p);
}

/*==========================================================================*/
char *mm_zloc(BUFFER *b, int item_len)    
/*+++
.PURPOSE Finds the position for a new (free) item. The found
	position is filled with zeroes, and b->offset is initialized.
.RETURNS Address of found item position / NULL if impossible.
.REMARKS No tracing.
  [IN] b        = The buffer to use
  [IN] item_len = Size of 1 item
---*/
{
  char	*p, *e;

    b->offset = b->used;	/* Default: at end of buffer */

    if ((b->offset + item_len) > b->allocated) {   /* No space left;
					Find a zero-filled position */
	for (b->offset = 0; b->offset < b->used; b->offset += item_len) {
	    p = b->buf + b->offset; e = p + item_len;
	    while ((p < e) && (*p == '\0')) p++;
	    if (p == e) break;	
	}
	if (b->offset >= b->used) /* Last possibility: expand buffer... */
    	    mm_balloc(b, item_len);
	 	/* Remember that b->offset is correctly set to b->used  */
    }

	/* Now, initialize the found item	*/

    if ((b->offset + item_len) <= b->allocated) {	/* It's OK...	*/
  	p = b->buf + b->offset;
  	bzero(b->buf + b->offset, item_len);
	if (b->used <= b->offset) b->used = b->offset + item_len;
    }
    else p = (char *)0;		/* No more space ... */

    return(p);
}

#include <stdarg.h>
char *mm_printf(BUFFER *buffer, char *format, ...)
/*++++++++++++++++
.PURPOSE  Append the diagnostic to the buffer buffer
.RETURNS  1 when str starts by beg
-----------------*/
{
  va_list argv;
  int bytes, len, pos;
  char *b;
    va_start(argv, format);
    bytes = (strlen(format)|255)+1;	/* At least 256bytes */
    pos = buffer->used;
    b = mm_balloc(buffer, -bytes);	/* Reserve room */
    bytes = buffer->allocated - pos; 	/* actual number of available bytes */
    len = vsnprintf(b, bytes, format, argv);
    if (len>=bytes) {
	fprintf(stderr, "#+++Very long message (>%d):\n", bytes);
	vfprintf(stderr, format, argv);
	bytes = (len|255)+1;
	b = BUF_ReserveBytes(buffer, bytes);
	len = vsnprintf(b, bytes, format, argv);
    }
    /* Mark the edited text as used... */
    buffer->used += len;
    return(b);
}
